if [ -z ${CONDA_BUILD+x} ]; then
	source /opt/concourse/worker/volumes/live/f973604f-cb2d-4d28-4ade-01e7244d47e7/volume/pymongo_1561563846214/work/build_env_setup.sh
fi
python setup.py install --single-version-externally-managed --record=record.txt